(:~
 : This module contains the transformation and complementary functions of the
 : interactive view for the KMG website. Its main function, digital:transform,
 : is triggered by xml-utils:make-html-digital-view.
 :
 : @author Simon Sendler
 : TODO: cleanup namespaces and imports
 : TODO: update namespaces to something project-specific
 :)

xquery version "3.1";

module namespace digital="http://bdn-edition.de/ns/digital";

import module namespace bdn-utils="http://bdn-edition.de/ns/utils" at "utils.xqm";
import module namespace commons="http://bdn-edition.de/ns/commons" at "commons.xqm";
import module namespace config="https://sade.textgrid.de/ns/config" at "../config.xqm";
import module namespace single="http://bdn-edition.de/ns/single" at "single.xqm";

declare namespace tei="http://www.tei-c.org/ns/1.0";
declare namespace xhtml="http://www.w3.org/1999/xhtml";
declare namespace output ="http://www.w3.org/2010/xslt-xquery-serialization";

(: for color palette cf. http://colorpalettes.net/color-palette-3653/  and
 : http://colorpalettes.net/color-palette-1481/ plus custom colors :)
declare variable $digital:colors :=
  map {
    "A": "#359d93",
    "a": "#359d93",
    "ab":  "#533380",
    "B":  "#823066",
    "b":  "#823066",
    "c":  "#df6b6a",
    "d":  "#f4b771",
    "e":  "#abd1e8",
    "f":  "#43640b",
    "z":  "#afca54",
    "bc": "#ded96f",
    "bd": "#3e5e04",
    "other": "#a4bfd1",
    "quote": "#79a6d2"
  };

(:~
 : The main transformation routine. It iterates all nodes of a given TEI/XML
 : file in the BdN specific intermediate format and procudes the HTML output for
 : the digital view of the KMG edition.
 :
 : This function is called by xml-utils:make-html-digital-view.
 :
 : For easier findability, elements in the typeswitch are sorted alphabetically.
 :
 : @author Simon Sendler
 : @version 1.0
 : @see http://bdn-edition.de/ns/digital
 : @param $nodes the nodes of a given TEI/XML document
 : @param $filename e.g. "2228v.xml"
 : @return the respective HTML nodes
 :)
declare function digital:transform($nodes as node()*, $filename as xs:string)
as node()*
{
  for $node in $nodes
  return
    (if ($node/@xml:id) then
      commons:make-text-anchor(string($node/@xml:id))
    else
      (),
    try
    {typeswitch ($node)

      case text() return
        text{replace($node, "@", " ")}

      case comment() return
        ()

        case element (tei:app) return
            if ($node/ancestor::tei:rdg) then 
                ()
            else
                let $readings := $node/child::tei:rdg
                let $readings-no := count($readings)
            return
                (digital:transform($node/tei:lem, $filename),
                for $reading in $readings return
                    (digital:make-reading($reading, $filename),
                    if ($readings-no gt 1 and $reading = $readings[last()]) then
                        ()
                    else
                        <div class="rdg-digital-divider"/>
                    )
                )

        case element (tei:back) return
            ()

        case element (tei:bibl) return
            let $id := substring-after($node/@corresp, ":")
            return
                (digital:transform($node/node(), $filename),
                element i {
                    if ($node/@type = "KMG") then
                        (attribute onclick {"event.stopPropagation();openrefModal('"||$id||"', 'KMG')"},
                        attribute class {"fas fa-book-medical"})
                    else
                        (attribute onclick {"event.stopPropagation();openrefModal('"||$id||"', 'base')"},
                        attribute class {"fas fa-book"})
                })

      case element (tei:choice) return
        let $id := util:hash(generate-id($node/tei:sic), "md5")
        let $target := "#" || $id
        let $icon := "fas fa-pen"
        return
            (element xhtml:span {
                if ($node/tei:sic/node()) then
                    digital:transform($node/tei:sic, $filename)
                else
                    <i>&#xF8;</i>,
                element i {
                    attribute onclick {"event.stopPropagation();$('#corrModal').modal('toggle');setCorrModalContent('"||$id||"')"},
                    attribute class {$icon}
                }
            },
            element xhtml:div {
                attribute id {$id},
                attribute style {"display:none"},
                digital:transform($node/tei:corr, $filename),
                " [" 
                || digital:resolve-name($node/tei:corr), 
                if ($node/tei:corr/@resp) then
                    let $id := substring-after(digital:resolve-reference($node/tei:corr), ":")
                        return
                            element i {
                                attribute onclick {"event.stopPropagation();openpersonModal('" || $id || "')"},
                                attribute class {"fas fa-user"}
                            }
                else
                    (),
                "]"
            }
            )

      case element (tei:corr) return
        if ($node/text()) then
          digital:transform($node/node(), $filename)
        else
          text {"&#xF8;"}

      case element (tei:div) return
       element xhtml:div {
         attribute {"type"} {$node/@type},
         digital:transform($node/node(), $filename)
       }

        case element (tei:figure) return
            element figure {
                attribute class {"image"},
                digital:transform($node/node(), $filename)
            }

        case element (tei:front) return
            element xhtml:div {
                attribute class {"preface"},
                digital:transform($node/node(), $filename)
            }

        case element (tei:graphic) return
            let $target := replace($node/@url, "textgrid:", "https://staging.mollenhauer-edition.de/exist/apps/sade/textgrid/data/kmg-images/") || ".jpg"
            return
                element xhtml:a {
                    attribute href {$target},
                    attribute target {"_blank"},
                    element xhtml:img {
                        attribute src {$target},
                        attribute alt {digital:transform($node/ancestor::tei:figure//tei:figDesc, $filename)}
                    }
                }

        case element (tei:head) return
            if ($node/@type="ToC") then
                ()
            else if ($node/parent::tei:div/@type = "part") then
                element xhtml:h2 {
                    digital:transform($node/node(), $filename)
                }
            else if ($node/parent::tei:div/@type = "chapter") then
                element xhtml:h3 {
                    digital:transform($node/node(), $filename)
                }
            else if ($node/@type="sub") then
                element xhtml:h3 {
                    attribute class {"sub-head"},
                    digital:transform($node/node(), $filename)
                }
            else if ($node/parent::tei:div/@type = "section") then
                element xhtml:h4 {
                    digital:transform($node/node(), $filename)
                }
            else
                element xhtml:h5 {
                digital:transform($node/node(), $filename)
                }

      case element (tei:hi) return
        element xhtml:span {
          attribute class {$node/@rend},
          digital:transform($node/node(), $filename)
        }

      case element (tei:item) return
        digital:make-list-item($node, $filename)

      case element (tei:l) return
        element xhtml:span {
          attribute class {"line"},
          digital:transform($node/node(), $filename),
          if ($node/following-sibling::tei:l) then
            element xhtml:br {}
          else
            ()
        }

        case element (tei:label) return
            switch ($node/@type)
                case "head" return
                    element xhtml:span {
                        attribute class {"headlabel"},
                        digital:transform($node/node(), $filename)
                    }
                case "list" return
                    ()
                case "note_anchor" return
                    element xhtml:span {
                        attribute class {"note_anchor"},
                        digital:transform($node/node(), $filename),
                        " "
                    }
                default return
                    digital:transform($node/node(), $filename)

        case element (tei:lb) return
            element xhtml:br {}

      case element (tei:lem) return
        if ($node/following-sibling::tei:rdg//tei:join or $node/following-sibling::tei:rdg//*/@copyOf) then 
            digital:transform($node/node(), $filename) 
        else
            let $reading := $node/following-sibling::tei:*
            let $reading-wits := string-join($reading/@wit, "") => replace("[# ]+", "")
            let $id := util:hash(generate-id($node/parent::*), "md5")
            return
                if ($node/node()) then
                    if ($node/tei:join) then
                        (element xhtml:div {
                            attribute class {"lem lem-block"},
                            attribute id {$id},
                            if ($reading-wits) then
                                attribute wits {$reading-wits}
                            else
                                (),
                            digital:transform(digital:resolve-join($node/tei:join), $filename)
                        })
                    else if ($node//tei:div or $node//tei:p or $node//tei:lg or $node//tei:list or $node//tei:figure or $node//tei:graphic or $node//tei:note) then
                        (element xhtml:div {
                            attribute class {"lem lem-block"},
                            attribute id {$id},
                            if ($reading-wits) then
                                attribute wits {$reading-wits}
                            else
                                (),
                            digital:transform($node/node(), $filename)
                        })
                    else
                        (element xhtml:span {
                            attribute class {"lem"},
                            attribute id {$id},
                            if ($reading-wits) then
                                attribute wits {$reading-wits}
                            else
                                (),
                            digital:transform($node/node(), $filename)
                        })
                else
                    element xhtml:span {
                        attribute class {"lem insertion"},
                        attribute id {$id},
                        if ($reading-wits) then
                            attribute wits {$reading-wits}
                        else
                            (),
                        <sup>{$reading-wits}</sup>,
                        text {"&#x221A;"}
                    }

      case element (tei:lg) return
        element xhtml:div {
          attribute class {"linegroup"},
          digital:transform($node/node(), $filename)
        }

      case element (tei:list) return
        element xhtml:ul {
          (if ($node/parent::tei:item) then
              attribute class {"nested-list labeled_list" ||
              (if ($node/@rend= "blockquote") then " blockquote"
                else if ($node/@rend= "blockquoteitalics") then " blockquote italics"
                else if ($node/@rend = "small_letters") then " small-letters"
                else "")}
            else if ($node/ancestor::tei:rdg) then
              attribute class {"rdg-list labeled_list" ||
                (if ($node/@rend= "blockquote") then " blockquote"
                else if ($node/@rend= "blockquoteitalics") then " blockquote italics"
                else if ($node/@rend = "small_letters") then " small-letters"
                else "")}
            else
              attribute class {"text-list labeled_list" ||
                (if ($node/@rend= "blockquote") then " blockquote"
                else if ($node/@rend= "blockquoteitalics") then " blockquote italics"
                else if ($node/@rend = "small_letters") then " small-letters"
                else "")}),
            digital:transform($node/node(), $filename)
        }

        case element (tei:milestone) return
            if ($node/@type = "paragraph") then
                element xhtml:br {}
            else
                ()

      case element (tei:name) return
        let $type := $node/@type
        let $classMap :=
            map {
                "art": "artname",
                "brandname": "brandname",
                "music": "musicname",
                "org": "orgname",
                "periodica": "periodname",
                "person": "persname"
            }
        let $iconMap :=
            map {
                "art": "far fa-image",
                "brandname": "far fa-copyright",
                "music": "fas fa-music",
                "org": "fas fa-briefcase",
                "periodica": "fas fa-newspaper",
                "person": "fas fa-user"
            }

        return
            element xhtml:span {
                attribute class {$classMap($type)},
                digital:transform($node/node(), $filename),
                if ($node/@ref) then
                    if (starts-with($node/@ref, "#")) then
                        let $id := substring-after(digital:resolve-reference($node), ":")
                        return
                            element i {
                                attribute onclick {"event.stopPropagation();open" || $type || "Modal('" || $id || "')"},
                                attribute class {$iconMap($type)}
                            }
                    else
                        let $id := substring-after($node/@ref, ":")
                        return
                            element i {
                                attribute onclick {"event.stopPropagation();open" || $type || "Modal('" || $id || "')"},
                                attribute class {$iconMap($type)}
                            }
                else
                    ()
            }

      case element (tei:note) return
            if ($node/@type="KMG" or $node/@type="editorial") then
                let $id := util:hash(generate-id($node), "md5")
                let $target := "#" || $id
                let $icon := "fas fa-comment-alt"
                return
                    (element xhtml:span {
                        element i {
                            attribute onclick {"event.stopPropagation();$('#noteModal').modal('toggle');setNoteModalContent('"||$id||"')"},
                            attribute class {$icon}
                        }
                    },
                    element xhtml:div {
                        attribute id {$id},
                        attribute style {"display:none"},
                        digital:transform($node/node(), $filename)
                    }
                    )
            else
                if ($node/ancestor::tei:p) then
                    element xhtml:span {
                        attribute class {"note_anchor"},
                        if ($node//tei:app//tei:label[@type="note_anchor"]) then
                            digital:transform($node//tei:app[descendant::tei:label[@type="note_anchor"]], $filename)
                        else
                            $node//tei:label[@type="note_anchor"]
                    }
                else
                    element xhtml:div {
                        attribute class {"note"},
                        digital:transform($node/node(), $filename)
                    }

        case element (tei:p) return
            (if ($node/ancestor::tei:figure) then
                element xhtml:figcaption {
                    digital:transform($node/node(), $filename)
                }
            else
                element xhtml:div {
                    attribute class {
                        if ($node/parent::tei:item/ancestor::tei:list) then
                            "p listp"
                        else if ($node/@rend="blockquoteitalics") then
                            "p italics"
                        else if ($node/@rend) then
                            $node/@rend
                        else
                            "p"
                    },
                    digital:transform($node/node(), $filename)
                },
            for $note in $node//tei:note[not(@type)] return
                element xhtml:div {
                    attribute class {"note"},
                    digital:transform($note/node(), $filename)
                })

      case element (tei:pb) return
        element xhtml:span {
          attribute class {"pb"},
          "|" || replace($node/@edRef, "#", "") || $node/@n || "|"
        }

        case element (tei:q) return 
            digital:make-quote($node, $filename)

        case element (tei:quote) return 
            digital:make-quote($node, $filename)

        case element (tei:ref) return
            element xhtml:span {
                attribute class {"ref"},
                digital:transform($node/node(), $filename),
                element xhtml:a {
                    attribute class {"text-link"},
                    attribute href {
                        if (starts-with($node/@target, "#")) then
                            $node/@target
                        else
                            replace($node/@target, "textgrid:", "/exist/apps/sade/textgrid/data/kmg/")
                    },
                    element xhtml:i {
                        attribute class {"fas fa-link"}
                    }
                }
            }

        case element (tei:seg) return
            let $id := util:hash(generate-id($node/tei:sic), "md5")
            let $target := "#" || $id
            let $icon := "fas fa-pen"
            return
                if ($node/@type = "quote_add" and $node/@resp) then
                    (element xhtml:span {
                        digital:transform($node/node(), $filename),
                        element xhtml:i {
                            attribute onclick {"event.stopPropagation();$('#corrModal').modal('toggle');setCorrModalContent('"||$id||"')"},
                            attribute class {$icon}
                        },
                        element xhtml:div {
                            attribute id {$id},
                            attribute style {"display:none"},
                            "Hinzufügung von",
                            digital:resolve-name($node),
                            let $id := substring-after(digital:resolve-reference($node), ":")
                            return
                                element i {
                                    attribute onclick {"event.stopPropagation();openpersonModal('" || $id || "')"},
                                    attribute class {"fas fa-user"}
                                },
                            "."
                        }
                    })
                else
                    digital:transform($node/node(), $filename)

        case element (tei:teiHeader) return
            ()

        case element (tei:text) return
            if ($node/@type="commentary") then ()
            else digital:transform($node/node(), $filename)

        default return
            digital:transform($node/node(), $filename)
    } catch * {<xhtml:span style="color:red;display:block">Error: {$err:code}. {$err:description} <br/> Offending node: {$node/name()}</xhtml:span>})
};

(:~
 : Creates the correct kind of quotation marks around a quote
 :
 : This function is called by digital:transform
 :
 : @author Simon Sendler
 : @version 1.0
 : @param $node A tei-Element containing a quotation
 : @param $filename E.g. "1234v.xml"
 : @return A xhtml:q
 :)

declare function digital:make-quote($node as node(), $filename as xs:string)
as element(xhtml:q)
{
    element xhtml:q
        {
            if ($node/@type) then attribute type {$node/@type} else (),
            switch ($node/@rend)
                case "double" return
                    ("&#x201E;", digital:transform($node/node(), $filename), "&#x201C;")
                case "single" return
                    ("&#x201E;", digital:transform($node/node(), $filename), "&#x201C;")
                case "qm_low" return
                    ("&#x201E;", digital:transform($node/node(), $filename), "&#x201C;")
                case "qm_high" return
                    ("&#x201C;", digital:transform($node/node(), $filename), "&#x201D;")
                case "g_out" return
                    ("&#xAB;", digital:transform($node/node(), $filename), "&#xBB;")
                case "g_in" return
                    ("&#xBB;", digital:transform($node/node(), $filename), "&#xAB;")
                default return
                    digital:transform($node/node(), $filename)
    }
};

(:~
 : Combines a tei:item and the preceding tei:label into one
 : xhtml:li
 :
 : This function is called by digital:transform
 :
 : @author Simon Sendler
 : @version 1.0
 : @param $node A tei:item
 : @param $filename E.g. "1234v.xml
 : @return A xhtml:li
 :)

declare function digital:make-list-item($node as node(), $filename as xs:string)
as node()*
{
  element xhtml:li {
    if ($node/@xml:id) then
      attribute id {$node/@xml:id}
    else (),
    attribute class {"labeled_item"},
    element xhtml:span {
      attribute class {"list-label"},
      $node/preceding-sibling::tei:label[1]/text()
    },
    digital:transform($node/node(), $filename)
  }
};

(:~
 : Transforms a tei:rdg into a xhtml:span containing the reading, links it
 : to the corresponding lemma, and creates a corresponding xhtml:path
 :
 : The function xml-utils:make-html-digital-view will move xhtml:span into the
 : div[@id="variants"] and the xhtml:path to the SVG-layer during further
 : processing
 :
 : This function is called by digital:transform
 :
 : @author Simon Sendler
 : @version 1.0
 : @see
 : @param $reading A tei:rdg
 : @param $filename E.g. "1234v.xml"
 : @return A sequence of xhtml:span and xhtml:path
 :)

declare function digital:make-reading($reading as node(), $filename as xs:string)
as element()*
{ if ($reading//tei:join or $reading//*/@copyOf) then () else
  let $wits := replace($reading/@wit, "[#]+", "")
  let $color := digital:get-rdg-color($wits)
  let $wits := replace($wits, " ", "")

  let $reading-id := util:hash(generate-id($reading), "md5")
  let $lemma-id := util:hash(generate-id($reading/parent::*), "md5")
  let $target-lemma := "#" || $lemma-id
  let $target-reading := "#" || $reading-id
  return
    (element xhtml:span
    {
      attribute class {"rdg-digital rdg-" || $wits},
      attribute id {$reading-id},
      attribute target-lem {$target-lemma},
      attribute wits {$wits},
      element table
      {
        <tr>
          <td class="siglum rdg-table">
            <strong>{$wits}</strong>
          </td>
          <td class="rdg-table">
            {if ($reading/node()) then
              digital:transform($reading/node(), $filename)
            else
              <i>&#xF8;</i>
            }
          </td>
        </tr>
      }
    },
    element xhtml:path
    {
      attribute class {"rdg-path"},
      attribute d {"M10 10"},
      attribute target-rdg {$target-reading},
      attribute rdg-wits {$wits},
      attribute target-lem {$target-lemma},
      attribute stroke {$color},
      attribute fill {"transparent"}
    })
};

(:~
 : Retrieves the color that belongs to one or several witnesses of a tei:rdg.
 :
 : This function is called by digital:transform.
 :
 : @author Michelle Weidling
 : @version 1.0
 : @param $wits tei:rdg/@wit
 : @return the hex color codes a witness
 :)

declare function digital:get-rdg-color($wits as xs:string*)
as xs:string
{
    if (count(tokenize($wits, " ")) gt 1) then
        let $wit-combination :=
            tokenize($wits, " ")
            => string-join("")
        let $color-hex := map:get($digital:colors, $wit-combination)
        return
            if ($color-hex) then
                $color-hex
            else
                map:get($digital:colors, "other")
    else
      map:get($digital:colors, $wits)
};

(:~
 : Resolves an abbreviated reference, adding a base-url to an id
 :
 : @author Simon Sendler
 : @version 1.0
 : @param $node A node containing an @corresp, @ref, @target, or @resp
 : @return A string containing the full reference
 :)

declare function digital:resolve-reference($node as node())
as xs:string?
{
  $node/root()//*[@xml:id = replace($node/attribute()[name() = ('corresp', 'ref', 'target', 'resp')], '#', '')]/@ref
};

(:~
 : Retrieves a name that is referenced elsewhere in the document
 :
 : @author Simon Sendler
 : @version 1.0
 : @param $node A node containing an @corresp, @ref, @target, or @resp
 : @return A string containing the resolved name
 :)

declare function digital:resolve-name($node as node())
as xs:string?
{
  string-join(
    $node/root()//*[@xml:id = replace($node/attribute()[name() = ('corresp', 'ref', 'target', 'resp')], '#', '')]/text()
  )
};

(:~
 : Retrieves the target node of a given @copyOf-reference.
 :
 : @author Michelle Weidling
 : @version 1.0
 : @param $node The current node
 : @param $copyOf A reference to another node
 : @return The node the @copyOf refers to
 :)
declare function digital:find-corresp-node($node as node()*, $copyOf as xs:string)
as node()?
{
  let $copyOf := replace($copyOf, "#", "")
  return
    $node/root()//*[@xml:id = $copyOf]
};

declare function digital:resolve-join($node as node()) as node()? {
    let $result := "tei:" || $node/@result
    let $targets := replace($node/@target, "#", "") => tokenize()
    return
        element {$result} {
            for $target in $targets return
                $node/ancestor::tei:app/tei:rdg//*[@xml:id = $target]/node()
        }
};